#include "Burden.h"
#include "Constant.h"
#include "math.h"
#include <string>
#include <cstring>
#include <iostream>
#include <stdlib.h>
#include "qfc.h"
#include "MixChidist.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>
#include <Eigen/SVD>

String Burden::genename = "";
String Burden::genename_f = "";
String Burden::outputFile = "";
double Burden::mafMin = 0.0;
double Burden::mafMax = 0.05;
double Burden::MAC = 0.0;
double Burden::miss = 0.05;
String Burden::chromosome = "";
bool Burden::vcfAnnotated = true;
bool Burden::CMC_binary = false;
bool Burden::CMC_counts = false;
bool Burden::MB = false;
bool Burden::burden = true;
bool Burden::LinScoreAssoc = false;
bool Burden::SCORE=true;
bool Burden::LRT = false;
bool Burden::WALD = false;
bool Burden::SKAT_BETA = false;
bool Burden::SKAT_UW = false;
bool Burden::SKAT_MB = false;
bool Burden::VT = false;
bool Burden::fullResult = true;

Burden::Burden(){}

Burden::~Burden()
{}

bool Burden::GetGenotypeVectorDosage(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum) 
{
   bool status = false;
   varsIncluded = "";

   //mac is the total number of alternative alleles included in the group
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   //Eigen::VectorXf mafSqrIncluded;

   for(int c=0;c<group.uniqueChr[groupNum].Length();c++)
   {
      reader.set1BasedReadSection(group.uniqueChr[groupNum][c].c_str(),group.startPos[groupNum][c],group.endPos[groupNum][c]+1);
      while(reader.readRecord(record))
      {
         int m;
         int position = record.get1BasedPosition();
         String altAllele = record.getAltStr();
         String refAllele = record.getRefStr();
         bool match = false;
         for(int j=0;j<group.chrList[groupNum].Length();j++)
         {
            if(position == group.posList[groupNum][j])
            {
               if(group.uniqueChr[groupNum][c] == group.chrList[groupNum][j])
               {
                  match=true;
                  m = j;
                  break;
               }
            }
         }
         if(!match)
            continue;

	 String allele = refAllele + "/" + altAllele;
	 if(allele != group.alleleList[groupNum][m])
	 {
	    String flipAllele = altAllele + "/" + refAllele;
	    if(flipAllele != group.alleleList[groupNum][m])
	    {
	       continue;
	    }
	    printf("ref and alt alleles from group files are flipped from vcf file!\n");
	 }

	 int n=0,nf=0;
	 double fmaf = 0.0, maf=0.0;
	 VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

	 bool XStatus = false;
	 if(AutoFit::fitX)
	 {
	    String chr = group.chrList[groupNum][0].c_str();
	    if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }

	 Eigen::VectorXf genotype;   
	 genotype.resize(trans.persons);
	 genotype.setConstant(-1.0);

	 //Loop through samples according to trans.pPheno 
	 //read in dosage into genotype vector
	 int idx=0;
	 double missing=0.0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int j=0;j<trans.pPheno[f].Length();j++)
	    {
	       String sample = ped[trans.pPheno[f][j]].pid;

	       //get the sample number by sample ID from hash
	       int s = trans.sampleVCFIDHash.Integer(sample);
	       if(s!=-1)
	       {
		  const std::string * geno = genoInfo.getString("DS",s);
		  //if dosage is not missing
		  if(strcmp((*geno).c_str(),".")>0)
		  {
		     float dose = atof((*geno).c_str());
		     if(dose >=0)
		     {
			if(XStatus && dose >0 && ped[trans.pPheno[f][j]].sex==SVT::maleLabel)
			   genotype[idx] = 2;
			else
			   genotype[idx] = dose;

			maf += genotype[idx];
			n++;

			//judge if this is a founder
			int founder = trans.foundersHash.Integer(sample);
			if(founder != -1)
			{
			   fmaf += genotype[idx];
			   nf++;
			}
		     }
		  }
		  else
		     missing++;
	       }
	       else
	       {
		  error("FATAL ERROR! Sample %s is not included in VCF file.\n",sample.c_str());
	       }
	       idx++;
	    }
	 }
	 maf /= 2.0*n;
	 fmaf /= 2.0*nf;

	 if(fmaf==0.0 && maf>0)
	    fmaf = 1.0/(2.0*nf);

	 missing /= trans.persons;
	 if(fmaf==0.0 || fmaf==1.0 || maf==0.0 || maf==1.0)
	    continue;
	 //check if frequencies are read from outside
	 if(!group.freqOut)
	 {
	    //check if maf falls into the range
	    if(SVT::FounderFreq)
	    {
	       double fmaf_tmp;
	       fmaf_tmp = fmaf;
	       if(fmaf>0.5)
		  fmaf_tmp = 1.0-fmaf;
	       if(fmaf_tmp>mafMax || fmaf_tmp<mafMin)
		  continue;
	    }
	    else
	    {
	       double maf_tmp;
	       maf_tmp=maf;
	       if(maf>0.5)
		  maf_tmp = 1.0-maf;
	       if(maf_tmp>mafMax || maf_tmp<mafMin)
		  continue;
	    }
	 }

	 if(missing>miss)
	    continue;

	 if(maf>0.5 && maf>1.0-mafMax)
	 {
	    status=true;
	 }
	 else if(maf<0.5 && maf<mafMax)
	 {
	    status = true;
	 }
	 else
	 {
	    continue;
	 }
	 numVarPerGene++;

	 bool flip=false;
	 if(SVT::FounderFreq)
	 {
	    if(fmaf>0.5)
	    {
	       flip=true;
	    }
	 }
	 else if(maf>0.5)
	 {
	    flip=true;
	 }

	 if(flip)
	 {
	    fmaf = 1.0-fmaf;
	    maf = 1.0-maf;
	    for(int i=0;i<trans.persons;i++)
	    {
	       if(genotype[i]!=-1.0)
	       {
		  genotype[i] = 2.0-genotype[i];
	       }
	    }
	 }

	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       mac+=genotype[i];
	    }
	 }

	 if(fullResult)
	 {
	    String tmp = "";
	    tmp += group.chrList[groupNum][m-1];
	    tmp+=":";
	    tmp+=group.posList[groupNum][m-1];
	    tmp+= "_";
	    if(flip)
	    {
	       tmp+=altAllele;
	       tmp+="/";
	       tmp += refAllele;
	    }
	    else
	    {
	       tmp+=refAllele;
	       tmp+="/";
	       tmp += altAllele;
	    }
	    tmp+=  ";";
	    varsIncluded += tmp;
	 }

	 if(SVT::FounderFreq)
	 {
	    mafIncluded.Push(fmaf);
	 }
	 else 
	 {
	    mafIncluded.Push(maf);
	 }

	 //update genotype vector and genotype matrix
	 if(SKAT_BETA || SKAT_UW || SKAT_MB || fullResult)
	 {
	    genotypeSKAT.conservativeResize(trans.persons,numVarPerGene);
	    for(int row=0;row<trans.persons;row++)
	    {
	       genotypeSKAT(row,numVarPerGene-1) = genotype[row];
	    }
	 }

	 if(CMC_counts)
	 {
	    for(int row=0;row<trans.persons;row++)
	    {
	       if(genotype[row]==-1.0)
		  genotypeCMC_c[row] += 2.0*maf;
	       else
		  genotypeCMC_c[row] += genotype[row];
	    }
	 }

	 if(CMC_binary)
	 {
	    for(int row=0;row<trans.persons;row++)
	    {
	       if(genotype[row]>0 && genotypeCMC_b[row]==0)
	       {
		  genotypeCMC_b[row] = 1.0;
	       }
	    }
	 }

	 if(MB)
	 {
	    for(int row=0;row<trans.persons;row++)
	    {
	       if(genotype[row]==-1.0)
	       {
		  if(SVT::FounderFreq)
		  {
		     genotypeMB[row] += 2.0*fmaf/sqrt(fmaf*(1.0-fmaf));
		  }
		  else
		  {
		     genotypeMB[row] += 2.0*maf/sqrt(maf*(1.0-maf));
		  }
	       }
	       else
	       {
		  if(SVT::FounderFreq)
		  {     
		     genotypeMB[row] += genotype[row]/sqrt(fmaf*(1.0-fmaf));
		  }     
		  else  
		  {     
		     genotypeMB[row] += genotype[row]/sqrt(maf*(1.0-maf));
		  }
	       }
	    }
	 }
      }
   }

   if(CMC_counts)
   {
      float mean = genotypeCMC_c.mean();
      AdjustMean(genotypeCMC_c,mean);
   }
   if(MB)
   {
      float mean = genotypeMB.mean();
      AdjustMean(genotypeMB,mean);
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   freq_min=mafIncluded.Min();
   freq_max=mafIncluded.Max();
   freq_mean=mafIncluded.Average();
   freq_se = mafIncluded.StandardDeviation();
   //printf("There are %d variants included in the gene\n",numVarPerGene);

   if(mac<MAC) 
      status = false;
   return status;
}

bool Burden::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   varsIncluded = "";
   //mac is the total number of alternative alleles included
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();
   Eigen::VectorXf mafSqrIncluded;

   for(int c=0;c<group.uniqueChr[groupNum].Length();c++)
   {
      reader.set1BasedReadSection(group.uniqueChr[groupNum][c].c_str(),group.startPos[groupNum][c],group.endPos[groupNum][c]+1);
      while(reader.readRecord(record))
      {
	 int m;
	 int position = record.get1BasedPosition();
	 String altAllele = record.getAltStr();
	 String refAllele = record.getRefStr();
	 bool match = false;
	 for(int j=0;j<group.chrList[groupNum].Length();j++)
	 {
	    if(position == group.posList[groupNum][j])
	    {
	       if(group.uniqueChr[groupNum][c] == group.chrList[groupNum][j])
	       {
		  match=true;
		  m = j;
		  break;
	       }
	    }
	 }
	 if(!match)
	    continue;

	 bool XStatus = false;

	 if(AutoFit::fitX)
	 {
	    String chr = group.chrList[groupNum][0].c_str();
	    if(chr==SVT::xLabel && position > SVT::Xstart && position < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }

	 String allele = refAllele + "/" + altAllele;
	 if(allele != group.alleleList[groupNum][m])
	 {  
	    String flipAllele = altAllele + "/" + refAllele;
	    if(flipAllele != group.alleleList[groupNum][m])
	    {  
	       continue;
	    }  
	    printf("ref and alt alleles from group files are flipped from vcf file!\n");
	 }

	 double fmaf=0.0,maf=0.0;
	 int n=0,nf=0;

	 //Fill in genotype vector
	 Eigen::VectorXf genotype;
	 genotype.resize(trans.persons);
	 genotype.setConstant(-1.0);

	 int idx=0;
	 double missing = 0.0;
	 for (int f = 0; f < ped.familyCount; f++)
	 {
	    for(int p=0;p<trans.pPheno[f].Length();p++)
	    {
	       String sample = ped[trans.pPheno[f][p]].pid.c_str();

	       //get the sample number by sample ID from hash
	       int s = trans.sampleVCFIDHash.Integer(sample);
	       if(s!=-1)
	       {
		  int numGTs = record.getNumGTs(s);
		  Vector allele;

		  for(int j = 0; j < numGTs; j++)
		  {
		     allele.Push(record.getGT(s,j));
		  }
		  int bad =0;
		  for(int m_=0;m_<allele.Length();m_++)
		  {
		     bad += (allele[m_]==VcfGenotypeSample::MISSING_GT);
		  }
		  //if no missing allele exists
		  //if no missing allele exists
		  if(bad==0)
		  {
		     int sum=0;
		     for(int i=0;i<allele.Length();i++)
			sum+=allele[i];

		     if(XStatus && sum >0 && ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
			genotype[idx] = 2;
		     else
			genotype[idx] = sum;

		     maf += genotype[idx];
		     n++;

		     //judge if this is a founder
		     int founder = trans.foundersHash.Integer(sample);
		     if(founder!=-1)
		     {
			nf++;
			fmaf+=genotype[idx];
		     }
		  }
		  else 
		     missing++;
	       }
	       idx++;
	    }
	 }

	 //Adjust genotype here
	 maf /= 2.0*n;
	 fmaf /=2.0*nf;

	 if(fmaf==0.0 && maf>0)
	    fmaf = 1.0/(2.0*nf);
	 missing /= trans.persons;
	 if(fmaf==0.0 || fmaf==1.0 || maf==0.0 || maf==1.0)
	    continue;

	 //check if maf falls into the range
	 if(!group.freqOut)
	 {
	    if(SVT::FounderFreq)
	    {
	       double fmaf_tmp;
	       fmaf_tmp = fmaf;
	       if(fmaf>0.5)
		  fmaf_tmp = 1.0-fmaf;
	       if(fmaf_tmp>mafMax || fmaf_tmp<mafMin)
		  continue;
	    }
	    else
	    {
	       double maf_tmp;
	       maf_tmp=maf;
	       if(maf>0.5)
		  maf_tmp = 1.0-maf;
	       if(maf_tmp>mafMax || maf_tmp<mafMin)
		  continue;
	    }
	 }

	 if(missing>miss)
	    continue;

	 if(maf>0.5 && maf>1.0-mafMax)
	 {
	    status=true;
	 }
	 else if(maf<0.5 && maf<mafMax)
	 {
	    status = true;
	 }
	 else
	 {
	    continue;
	 }
	 numVarPerGene++;

	 bool flip=false;
	 if(SVT::FounderFreq)
	 {
	    if(fmaf>0.5)
	    {
	       flip=true;
	    }
	 }
	 else if(maf>0.5)
	 {
	    flip=true;
	 }

	 if(flip)
	 {
	    fmaf = 1.0-fmaf;
	    maf = 1.0-maf;
	    for(int i=0;i<trans.persons;i++)
	    {
	       if(genotype[i]!=-1.0)
	       {
		  genotype[i] = 2.0-genotype[i];
	       }
	    }
	 }

	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {  
	       mac+=genotype[i];
	    }  
	 }

	 if(fullResult)
	 {
	    String tmp;
	    tmp += group.chrList[groupNum][m];
	    tmp+=":";
	    tmp+=group.posList[groupNum][m];
	    tmp+= "_";
	    if(flip)
	    {
	       tmp+=altAllele;
	       tmp+="/";
	       tmp += refAllele;
	    }
	    else
	    {
	       tmp+=refAllele;
	       tmp+="/";
	       tmp += altAllele;
	    }
	    tmp+=  ";";
	    varsIncluded += tmp;
	 }

	 if(SVT::FounderFreq)
	 {
	    mafIncluded.Push(fmaf);
	 }
	 else
	 {
	    mafIncluded.Push(maf);
	 }

	 //update genotype vector and genotype matrix
	 if(SKAT_BETA || SKAT_UW ||SKAT_MB || fullResult)
	 {
	    genotypeSKAT.conservativeResize(trans.persons,numVarPerGene);
	    for(int row=0;row<trans.persons;row++)
	    {
	       genotypeSKAT(row,numVarPerGene-1) = genotype[row];
	    }
	 }

	 if(CMC_counts)
	 {
	    for(int row=0;row<trans.persons;row++)
	    {
	       if(genotype[row]==-1.0)
		  genotypeCMC_c[row] += 2.0*maf;
	       else
		  genotypeCMC_c[row] += genotype[row];
	    }
	 }

	 if(CMC_binary)
	 {
	    for(int row=0;row<trans.persons;row++)
	    {
	       if(genotype[row]>0 && genotypeCMC_b[row]==0)
	       {
		  genotypeCMC_b[row] = 1.0;
	       }
	    }
	 }

	 if(MB)
	 {
	    for(int row=0;row<trans.persons;row++)
	    {
	       if(genotype[row]==-1.0)
	       {
		  if(SVT::FounderFreq)
		  {
		     genotypeMB[row] += 2.0*fmaf/sqrt(fmaf*(1.0-fmaf));
		  }
		  else
		  {
		     genotypeMB[row] += 2.0*maf/sqrt(maf*(1.0-maf));
		  }
	       }
	       else
	       {
		  if(SVT::FounderFreq)
		  {
		     genotypeMB[row] += genotype[row]/sqrt(fmaf*(1.0-fmaf));
		  }
		  else
		  {
		     genotypeMB[row] += genotype[row]/sqrt(maf*(1.0-maf));
		  }
	       }
	    }
	 }
      }
   }

   if(CMC_counts)
   {
      float mean = genotypeCMC_c.mean();
      AdjustMean(genotypeCMC_c,mean);
   }
   if(MB)
   {
      float mean = genotypeMB.mean();
      AdjustMean(genotypeMB,mean);
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }

   freq_min=mafIncluded.Min();
   freq_max=mafIncluded.Max();
   freq_mean=mafIncluded.Average();
   freq_se = mafIncluded.StandardDeviation();

   if(mac<MAC)
      status = false;
   return status;
}

bool Burden::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;
   varsIncluded = "";

   //mac is the total number of alternative alleles included in the group
   mac = 0.0;
   //numVarPerGene is the total number of variants included in this gene.
   numVarPerGene=0;

   mafIncluded.Clear();

   for(int m=0;m<group.chrList[groupNum].Length();m++)
   {
      String markerName = group.chrList[groupNum][m];
      markerName += ":";
      markerName += group.posList[groupNum][m];
      int markerid = ped.markerLookup.Integer(markerName);
      if(markerid==-1)
      {
	 continue;
      }
      int alleleCount = CountAlleles(ped, markerid);
      if(alleleCount!=2)
	 continue;

      String refAllele = ped.GetMarkerInfo(markerid)->alleleLabels[1];
      String altAllele = ped.GetMarkerInfo(markerid)->alleleLabels[2];

      String allele = refAllele + "/" + altAllele;
      if(allele != group.alleleList[groupNum][m])
      {
	 String flipAllele = altAllele + "/" + refAllele;
	 if(flipAllele != group.alleleList[groupNum][m])
	 {
	    continue;
	 }
      }

      //if marker m from group list is found then get the genotype vector
      bool XStatus = false;
      if(AutoFit::fitX)
      {
	 int found = markerName.FastFind(SVT::xLabel,0);
	 if(found!=-1)
	 {
	    if(group.posList[groupNum][m] > SVT::Xstart && group.posList[groupNum][m] < SVT::Xend)
	    {
	       XStatus=true;
	    }
	 }
      }

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      Eigen::VectorXf genotype;
      genotype.resize(trans.persons);
      genotype.setConstant(-1.0);

      //Loop through samples according to trans.pPheno 
      int idx=0;
      double missing = 0.0;
      for(int f=0;f<ped.familyCount;f++)
      {
	 for(int p=0;p<trans.pPheno[f].Length();p++)
	 {
	    if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	    {
	       idx++;
	       missing++;
	       continue;
	    }
	    if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	    {
	       if(XStatus)
	       {
		  if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
		  {
		     genotype[idx] = 1.0;
		  }
		  if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
		  {
		     genotype[idx] = 2.0;
		  }
	       }
	       else
	       {
		  genotype[idx] = 1.0;
	       }
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	    {
	       if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	       {
		  genotype[idx] = 2.0;
	       }
	       else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	       {
		  genotype[idx] = 0.0;
	       }
	    }
	    maf+=genotype[idx];
	    n++;
	    int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	    if(founder != -1)
	    {
	       fmaf += genotype[idx];
	       nf++;
	    }
	    idx++;
	 }
      }

      maf /= 2.0*n;
      fmaf /= 2.0*nf;

      if(fmaf==0.0 && maf>0)
	 fmaf = 1.0/(2.0*nf);
      missing /= trans.persons;
      if(fmaf==0.0 || fmaf==1.0 || maf==0.0 || maf==1.0)
	 continue;

      //check if maf falls into the range
      if(!group.freqOut)
      {
	 if(SVT::FounderFreq)
	 {
	    double fmaf_tmp;
	    fmaf_tmp = fmaf;
	    if(fmaf>0.5)
	       fmaf_tmp = 1.0-fmaf;
	    if(fmaf_tmp>mafMax || fmaf_tmp<mafMin)
	       continue;
	 }
	 else
	 {
	    double maf_tmp;
	    maf_tmp=maf;
	    if(maf>0.5)
	       maf_tmp = 1.0-maf;
	    if(maf_tmp>mafMax || maf_tmp<mafMin)
	       continue;
	 }
      }

      if(missing>miss)
	 continue;

      if(maf>0.5 && maf>1.0-mafMax)
      {
	 status=true;
      }
      else if(maf<0.5 && maf<mafMax)
      {
	 status = true;
      }
      else
      {
	 continue;
      }
      numVarPerGene++;

      bool flip=false;
      if(SVT::FounderFreq)
      {
	 if(fmaf>0.5)
	 {
	    flip=true;
	 }
      }
      else if(maf>0.5)
      {
	 flip=true;
      }

      if(flip)
      {
	 fmaf = 1.0-fmaf;
	 maf = 1.0-maf;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotype[i]!=-1.0)
	    {
	       genotype[i] = 2.0-genotype[i];
	    }
	 }
      }

      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    mac+=genotype[i];
	 }
      }

      if(fullResult)
      {
	 String tmp = "";
	 tmp += group.chrList[groupNum][m];
	 tmp+=":";
	 tmp+=group.posList[groupNum][m];
	 tmp+= "_";
	 if(flip)
	 {
	    tmp+=altAllele;
	    tmp+="/";
	    tmp += refAllele;
	 }
	 else
	 {
	    tmp+=refAllele;
	    tmp+="/";
	    tmp += altAllele;
	 }
	 tmp+=  ";";
	 varsIncluded += tmp;
      }

      if(SVT::FounderFreq)
      {
	 mafIncluded.Push(fmaf);
      }
      else
      {
	 mafIncluded.Push(maf);
      }

      //update genotype vector and genotype matrix
      if(SKAT_BETA || SKAT_UW ||SKAT_MB || fullResult)
      {
	 genotypeSKAT.conservativeResize(trans.persons,numVarPerGene);
	 for(int row=0;row<trans.persons;row++)
	 {
	    genotypeSKAT(row,numVarPerGene-1) = genotype[row];
	 }
      }

      if(CMC_counts)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	       genotypeCMC_c[row] += 2.0*maf;
	    else
	       genotypeCMC_c[row] += genotype[row];
	 }
      }

      if(CMC_binary)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]>0 && genotypeCMC_b[row]==0)
	    {
	       genotypeCMC_b[row] = 1.0;
	    }
	 }
      }

      if(MB)
      {
	 for(int row=0;row<trans.persons;row++)
	 {
	    if(genotype[row]==-1.0)
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += 2.0*fmaf/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += 2.0*maf/sqrt(maf*(1.0-maf));
	       }
	    }
	    else
	    {
	       if(SVT::FounderFreq)
	       {
		  genotypeMB[row] += genotype[row]/sqrt(fmaf*(1.0-fmaf));
	       }
	       else
	       {
		  genotypeMB[row] += genotype[row]/sqrt(maf*(1.0-maf));
	       }
	    }
	 }
      }
   }
   if(CMC_counts)
   {
      float mean = genotypeCMC_c.mean();
      AdjustMean(genotypeCMC_c,mean);
   }
   if(MB)
   {
      float mean = genotypeMB.mean();
      AdjustMean(genotypeMB,mean);
   }

   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   freq_min=mafIncluded.Min();
   freq_max=mafIncluded.Max();
   freq_mean=mafIncluded.Average();
   freq_se = mafIncluded.StandardDeviation();
   //printf("There are %d variants included in the gene\n",numVarPerGene);

   if(mac<MAC)
      status = false;
   return status;
}

void Burden::AdjustMean(Eigen::VectorXf & genotype,float & mean)
{
   for(int i=0;i<genotype.size();i++)
      genotype[i] -= mean;
}

bool Burden::GetGenotypeVector(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = true;

   if(group.chrList[groupNum].Length()==0)
   {
      status = false;
      return status;
   }

   if(MB)
   {
      genotypeMB.resize(trans.persons);
      genotypeMB.setZero();
   }
   if(CMC_binary)
   {
      genotypeCMC_b.resize(trans.persons);
      genotypeCMC_b.setZero();
   }
   if(CMC_counts)
   {
      genotypeCMC_c.resize(trans.persons);
      genotypeCMC_c.setZero();
   }
   if(SKAT_BETA || SKAT_UW ||SKAT_MB || fullResult)
   {
      genotypeSKAT.resize(trans.persons,1);
      genotypeSKAT.setConstant(-1.0);
   }
   if(SVT::genoFromPed)
      status = GetGenotypeVectorPED(trans,ped,group,groupNum);
   if(SVT::genoFromVCF)
      status = GetGenotypeVectorVCF(trans,ped,group,groupNum);
   if(SVT::dosage)
      status = GetGenotypeVectorDosage(trans,ped,group,groupNum);

   if(!status)
      return status;

   if(SKAT_BETA || SKAT_UW ||SKAT_MB || fullResult)
   {
      //adjust genotype matrix by mean
      for(int col=0;col<genotypeSKAT.cols();col++)
      {
	 float mean = 0.0;
	 int n_=0;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotypeSKAT(i,col)!=-1.0)
	    {
	       mean += genotypeSKAT(i,col);
	       n_++;
	    }
	 }
	 mean /= n_;
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(genotypeSKAT(i,col)==-1.0)
	       genotypeSKAT(i,col)=0.0;
	    else
	       genotypeSKAT(i,col) -= mean;
	 }
      }
   }

   if(!SKAT_BETA && !SKAT_UW && !SKAT_MB)
      return status;
   return status;
}

void Burden::Calculate(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group,FILE * log)
{
   int traitNum = engine.GetTraitID(ped,traitName.c_str(),log);
   FILE * MBScoreOut = NULL;
   FILE * CMC_bScoreOut = NULL;
   FILE * CMC_cScoreOut = NULL;
   FILE * MBLRTOut = NULL;
   FILE * CMC_bLRTOut = NULL;
   FILE * CMC_cLRTOut = NULL;
   FILE * MBWALDOut = NULL;
   FILE * CMC_bWALDOut = NULL;
   FILE * CMC_cWALDOut = NULL;
   FILE * MBLinScoreOut = NULL;
   FILE * CMC_bLinScoreOut = NULL;
   FILE * CMC_cLinScoreOut = NULL;
   FILE * SKATOut = NULL;
   FILE * SKATOut_UW = NULL;
   FILE * SKATOut_MB = NULL;

   if(SCORE)
   {
      if(MB) 
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.score.txt";
	 MBScoreOut = fopen(MBfilename,"wt");
	 PrintHeader(MBScoreOut ,trans,engine,traitName);
      }

      if(CMC_binary) 
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.score.txt";
	 CMC_bScoreOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bScoreOut ,trans,engine,traitName);
      }

      if(CMC_counts) 
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".burden.score.txt";
	 CMC_cScoreOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cScoreOut,trans,engine,traitName);
      }
   }

   if(WALD)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.wald.txt";
	 MBWALDOut = fopen(MBfilename,"wt");
	 PrintHeader(MBWALDOut ,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.wald.txt";
	 CMC_bWALDOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bWALDOut ,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.wald.txt";
	 CMC_cWALDOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cWALDOut,trans,engine,traitName);
      }
   }

   if(LinScoreAssoc)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.LinScore.txt";
	 MBLinScoreOut = fopen(MBfilename,"wt");
	 PrintHeader(MBLinScoreOut ,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.LinScore.txt";
	 CMC_bLinScoreOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bLinScoreOut ,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.LinScore.txt";
	 CMC_cLinScoreOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cLinScoreOut,trans,engine,traitName);
      }
   }

   if(LRT)
   {
      if(MB)
      {
	 String MBfilename;
	 MBfilename = outputFile + traitName + ".Madsen-Browning.LRT.txt";
	 MBLRTOut = fopen(MBfilename,"wt");
	 PrintHeader(MBLRTOut,trans,engine,traitName);
      }

      if(CMC_binary)
      {
	 String CMC_bfilename;
	 CMC_bfilename = outputFile + traitName + ".CMC_binary.LRT.txt";
	 CMC_bLRTOut = fopen(CMC_bfilename,"wt");
	 PrintHeader(CMC_bLRTOut,trans,engine,traitName);
      }

      if(CMC_counts)
      {
	 String CMC_cfilename;
	 CMC_cfilename = outputFile + traitName + ".CMC_counts.LRT.txt";
	 CMC_cLRTOut = fopen(CMC_cfilename,"wt");
	 PrintHeader(CMC_cLRTOut,trans,engine,traitName);
      }
   }

   if(SKAT_BETA)
   {
      String SKATfilename;
      SKATfilename = outputFile + traitName + ".skat.betaweight.result.txt";
      SKATOut = fopen(SKATfilename,"wt");
      PrintSKATHeader(SKATOut,trans,engine,traitName);
   }

   if(SKAT_UW)
   {
      String SKATfilename;
      SKATfilename = outputFile + traitName + ".skat.unitweight.result.txt";
      SKATOut_UW = fopen(SKATfilename,"wt");
      PrintSKATHeader(SKATOut_UW,trans,engine,traitName);
   }

   if(SKAT_MB)
   {
      String SKATfilename;
      SKATfilename = outputFile + traitName + ".skat.MBweight.result.txt";
      SKATOut_MB = fopen(SKATfilename,"wt");
      PrintSKATHeader(SKATOut_MB,trans,engine,traitName);
   }

   double L0;

   if(LRT)
   {
      if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
      {
	 L0=engine.EvaluateMulti(trans,false);
      }
      else
      {
	 L0 = engine.Evaluate(engine.deltaHat,trans,false);
	 //printf("L0 is:%g\n",L0);
      }
   }

   int bad =0;
   if(SVT::dosage || SVT::genoFromVCF)
   {
      reader.open(GroupFromAnnotation::vcfInput,header);
      reader.readVcfIndex();
   }

   for(int i=0;i<group.annoGroups.Length();i++)
   {
      if(i%1000==1 && i>1000)
      {
	 printf("  Finished analyzing %d genes.\n",((int)i/1000)*1000);
      }
      String geneName = group.annoGroups[i];
      //printf("now working on gene %s\n",geneName.c_str());
      if(GetGenotypeVector(trans,ped,group,i))
      {
	 //printf("working on gene %s\n",geneName.c_str());
	 if(fullResult)
	 {
	    CalculateSingleVarPvalue(trans,geneName,engine,genotypeSKAT,ped);
	 }
	 Eigen::VectorXf transGenoMB;
	 Eigen::VectorXf transGenoCMC_b;
	 Eigen::VectorXf transGenoCMC_c;

	 if(!FastFit::unrelated)
	 {
	    if(MB)
	       //   transGenoMB.Product(trans.transU,genotypeMB);
	       transGenoMB = trans.UD*genotypeMB;
	    if(CMC_binary)
	       //transGenoCMC_b.Product(trans.transU,genotypeCMC_b);
	       transGenoCMC_b = trans.UD *genotypeCMC_b;
	    if(CMC_counts)
	       //transGenoCMC_c.Product(trans.transU,genotypeCMC_c);
	       transGenoCMC_c = trans.UD* genotypeCMC_c;
	 }

	 if(LRT || WALD)
	 {
	    FastTransform  fullTrans;
	    FastFit fullEngine;
	    fullTrans.persons = trans.persons;
	    fullTrans.fixedEffects = trans.fixedEffects +1;

	    if(AutoFit::fitX)
	    {
	       //int covariateId = ped.GetCovariateID("MarkerTested");

	       fullTrans.pPheno = new IntArray [ped.familyCount];
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  fullTrans.pPheno[f].Dimension(0);
		  for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		  {
		     if(ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
		     {
			fullTrans.pPheno[f].Push(i);
		     }
		  }
	       }
	       int index =0;
	       for (int f = 0; f < ped.familyCount; f++)
	       {
		  if (fullTrans.pPheno[f].Length() > 0)
		  {
		     int count = fullTrans.pPheno[f].Length();
		     for(int i=0;i<count;i++)
		     {
			if(MB)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeMB[index];
			if(CMC_binary)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeCMC_b[index];
			if(CMC_counts)
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotypeCMC_c[index];
			index++;
		     }
		  }
	       }
	       printf("ped.covariateCount is%d\n",ped.covariateCount);

	       AutoFit fullAutoEngine;
	       fullAutoEngine.useCovariates = true;
	       fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
	       fullAutoEngine.fitX = AutoFit::fitX;

	       //Fit the alternative
	       fullAutoEngine.FitPolygenicModels(ped,traitNum,true);

	       fullTrans.Y.Copy(trans.Y);
	       fullTrans.X.Copy(trans.X);
	       fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
	       fullTrans.UY.Dimension(fullTrans.persons);
	       fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.transU.resize(0,0);
	       fullTrans.D.Dimension(0);

	       if(MB)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeMB[i];
	       if(CMC_binary)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeCMC_b[i];
	       if(CMC_counts)
		  for(int i=0;i<fullTrans.persons;i++)
		     fullTrans.X[i][trans.fixedEffects] = genotypeCMC_c[i];

	       fullTrans.traitVar = fullAutoEngine.variance;
	       fullTrans.TransformPedkinX(fullAutoEngine, ped);
	       fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
	    }
	    else 
	    {
	       fullTrans.D.Copy(trans.D);
	       fullTrans.UY.Copy(trans.UY);
	       fullTrans.transU = trans.transU;
	       fullTrans.UD.resize(fullTrans.persons,fullTrans.persons);

	       fullEngine.beta.Dimension(fullTrans.fixedEffects);
	       fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
	       fullTrans.UX.Copy(trans.UX);
	       fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
	    }

	    String method;
	    if(LRT)
	    {
	       if(MB)
	       {
		  method = "MB";
		  MerlinLRT(L0,fullTrans,MBLRTOut,transGenoMB,geneName,fullEngine,method);
	       }
	       if(CMC_binary)
	       {
		  method = "CMC_binary";
		  MerlinLRT(L0,fullTrans,CMC_bLRTOut,transGenoCMC_b,geneName,fullEngine,method);
	       }
	       if(CMC_counts)
	       {
		  method = "CMC_counts";
		  MerlinLRT(L0,fullTrans,CMC_cLRTOut,transGenoCMC_c,geneName,fullEngine,method);
	       }
	    }

	    if(WALD)
	    {
	       if(MB)
	       {
		  method = "MB";
		  WaldAssoc(fullTrans,MBWALDOut,transGenoMB,geneName,fullEngine,method);
	       }
	       if(CMC_binary)
	       {
		  method = "CMC_binary";	
		  WaldAssoc(fullTrans,CMC_bWALDOut,transGenoCMC_b,geneName,fullEngine,method);
	       }
	       if(CMC_counts)
	       {
		  method = "CMC_counts";	
		  WaldAssoc(fullTrans,CMC_cWALDOut,transGenoCMC_c,geneName,fullEngine,method);
	       }
	    }
	 }

	 String method;
	 if(LinScoreAssoc)
	 {
	    if(MB)
	    {
	       method = "MB";
	       LinAssoc(trans,engine,MBLinScoreOut,genotypeMB,geneName,method);
	    }
	    if(CMC_binary)
	    {
	       method = "CMC_binary";
	       LinAssoc(trans,engine,CMC_bLinScoreOut,genotypeCMC_b,geneName,method);
	    }
	    if(CMC_counts)
	    {
	       method = "CMC_counts";
	       LinAssoc(trans,engine,CMC_cLinScoreOut,genotypeCMC_c,geneName,method);
	    }
	 }

	 if(SCORE)
	 {
	    if(MB)
	    {
	       method = "MB";
	       if(FastFit::unrelated)
		  MerlinFastAssoc(trans,MBScoreOut,genotypeMB,geneName,engine,method);
	       else
		  MerlinFastAssoc(trans,MBScoreOut,transGenoMB,geneName,engine,method);
	    }
	    if(CMC_binary)
	    {
	       method = "CMC_binary";
	       if(FastFit::unrelated)
		  MerlinFastAssoc(trans,CMC_bScoreOut,genotypeCMC_b,geneName,engine,method);
	       else
		  MerlinFastAssoc(trans,CMC_bScoreOut,transGenoCMC_b,geneName,engine,method);
	    }
	    if(CMC_counts)
	    {
	       method = "CMC_counts";
	       if(FastFit::unrelated)
		  MerlinFastAssoc(trans,CMC_cScoreOut,genotypeCMC_c,geneName,engine,method);
	       else
		  MerlinFastAssoc(trans,CMC_cScoreOut,transGenoCMC_c,geneName,engine,method);
	    }
	 }
	 if(SKAT_BETA || SKAT_UW || SKAT_MB)
	 {
	    Eigen::MatrixXf transGeno;
	    if(FastFit::unrelated)
	    {
	       transGeno = genotypeSKAT;
	    }
	    else
	    {
	       if(!trans.empKin)
	       {
		  transGeno.resize(genotypeSKAT.rows(),genotypeSKAT.cols());
		  transGeno.setZero();
		  for(int m=0;m<genotypeSKAT.cols();m++)
		  {
		     int index=0;
		     for (int f = 0; f < ped.familyCount; f++)
		     {
			if (trans.pPheno[f].Length() > 0)
			{
			   int count = trans.pPheno[f].Length();
			   int block = index;
			   for(int i=0;i<count;i++)
			   {
			      for(int c=0;c<count;c++)
			      {
				 transGeno(index,m) += trans.UD(block+i,block+c)*genotypeSKAT(block+c,m);
			      }
			      index++;
			   }
			}
		     }
		  }
	       }
	       else 
	       {
		  transGeno = trans.UD * genotypeSKAT;
	       }
	    }

	    String method;
	    if(SKAT_BETA)
	    {
	       method= "beta";
	       SKATScoreAssoc(trans,SKATOut,geneName,engine,ped,transGeno,method);
	    }
	    if(SKAT_UW)
	    {
	       method = "uw";
	       SKATScoreAssoc(trans,SKATOut_UW,geneName,engine,ped,transGeno,method);
	    }
	    if(SKAT_MB)
	    {
	       method = "mb";
	       SKATScoreAssoc(trans,SKATOut_MB,geneName,engine,ped,transGeno,method);
	    }
	 }
      } 
      else 
      {
	 bad++;
	 //printf("bad gene is:%s\n",group.annoGroups[i].c_str());
	 fprintf(log,"Gene %s is not analyzed due to no qualified markers to include.\n",geneName.c_str());
	 fflush(log);
      }
   }

   if(SCORE)
   {
      if(MB)
	 PrintFooter(MBScoreOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bScoreOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cScoreOut,bad,group);
   }
   if(LinScoreAssoc)
   {
      if(MB)
	 PrintFooter(MBLinScoreOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bLinScoreOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cLinScoreOut,bad,group);
   }
   if(LRT)
   {
      if(MB)
	 PrintFooter(MBLRTOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bLRTOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cLRTOut,bad,group);
   }
   if(WALD)
   {
      if(MB)
	 PrintFooter(MBWALDOut,bad,group);
      if(CMC_binary)
	 PrintFooter(CMC_bWALDOut,bad,group);
      if(CMC_counts)
	 PrintFooter(CMC_cWALDOut,bad,group);
   }
   if(SKAT_BETA)
      PrintFooter(SKATOut,bad,group);
   if(SKAT_UW)
      PrintFooter(SKATOut_UW,bad,group);
   if(SKAT_MB)
      PrintFooter(SKATOut_MB,bad,group);

   if(MB)
   {
      if(SCORE) fclose(MBScoreOut);
      if(LinScoreAssoc) fclose(MBLinScoreOut);
      if(LRT) fclose(MBLRTOut);
      if(WALD) fclose(MBWALDOut);
   }
   if(CMC_binary)
   {
      if(SCORE) fclose(CMC_bScoreOut);
      if(LinScoreAssoc) fclose(CMC_bLinScoreOut);
      if(LRT) fclose(CMC_bLRTOut);
      if(WALD) fclose(CMC_bWALDOut);
   } 
   if(CMC_counts)
   {
      if(SCORE) fclose(CMC_cScoreOut);
      if(LinScoreAssoc) fclose(CMC_cLinScoreOut);
      if(LRT) fclose(CMC_cLRTOut);
      if(WALD) fclose(CMC_cWALDOut);
   }
   if(SKAT_BETA)
      fclose(SKATOut);
   if(SKAT_UW)
      fclose(SKATOut_UW);
   if(SKAT_MB)
      fclose(SKATOut_MB);
   //Restoring the pedigree covariate settings.
   if((AutoFit::fitX || AutoFit::fitSharedEnvironment) && (LRT || WALD))
   {
      ped.covariateNames.Delete(ped.covariateNames.Length()-1);
      ped.covariateLookup.Delete("MarkerTested");
      ped.covariateCount--;
      for(int i=0;i<ped.count;i++)
	 ped[i].covariates[ped.covariateCount] = 0.0;
   }
}

void Burden::Run(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group,FILE * log)
{
   if(!CMC_counts && !SKAT_BETA && !SKAT_UW && !SKAT_MB)
      return;
   printf("Performing ");
   if(CMC_counts)
   {
      printf("burden ");
   }
   if(SKAT_BETA || SKAT_UW || SKAT_MB)
   {
      if(burden)
	 printf("and SKAT ");
      else
	 printf("SKAT ");
   }
   printf("analyses ...\n");
   fflush(stdout);

   if(!vcfAnnotated)
   {
      error("\nERROR! VCF file needs to be annotated in INFO field.\n");
   }

   Eigen::VectorXf betaHat(engine.betaHat.Length());
   for(int i=0;i<engine.betaHat.Length();i++)
      betaHat[i] = engine.betaHat[i];

   residuals_original.resize(trans.persons);
   if(FastFit::unrelated)
   {
      fixedEff.resize(trans.persons);
      for(int i =0;i<trans.persons;i++)
	 fixedEff[i] = trans.X[i].InnerProduct(engine.betaHat);
      for(int i=0;i<trans.persons;i++)
	 residuals_original[i] = trans.Y[i] - fixedEff[i];
   }
   else
   {
      residuals_original = trans.UDY - trans.UDX*betaHat;
   }

   sigma2_hat =1.0;

   if(SKAT_BETA || SKAT_UW || SKAT_MB || VT)
   {
      P0.resize(trans.persons,trans.persons);
      P0 = trans.UDX * trans.inv * trans.UDX.transpose();
      for(int i=0;i<trans.persons;i++)
      {
	 P0(i,i) = 1.0 - P0(i,i);
	 //P0(i,i) *= sigma2_hat;
	 for(int j=i+1;j<trans.persons;j++)
	 {
	    P0(i,j) = -P0(i,j); 
	    P0(j,i) = P0(i,j); 
	 }
      }
   }

   Calculate(ped,trans,engine,traitName,group,log);
   printf("completed.\n");
}

void Burden::LinAssoc(FastTransform & trans,FastFit & engine,FILE * out, Eigen::VectorXf & genotype,String & geneName,String & method)
{
   //Fill in UDX, UD, and UDY
   //trans.FinalizeTransform(engine.sigma_g2Hat, engine.sigma_e2Hat);
   Eigen::VectorXf transGeno;
   transGeno = trans.UD * genotype;
   double pvalue,numerator,denominator,chisq,effSize;
   Eigen::VectorXf denominator_2;
   denominator_2.resize(trans.fixedEffects);
   numerator = 0.0;
   denominator = 1.0e-10;
   denominator_2.setZero();

   for(int i=0;i<trans.persons;i++)
   {
      numerator += residuals_original[i]*transGeno[i];
      denominator += transGeno[i]*transGeno[i];
      for(int j=0;j<trans.fixedEffects;j++)
      {
	 denominator_2[j] += transGeno[i]*trans.UDX(i,j);
      }
   }

   Eigen::VectorXf tmp; //tmp is T(sum_SiZi)*inv
   tmp = trans.inv*denominator_2;
   denominator -= tmp.dot(denominator_2);
   denominator *= sigma2_hat;

   /*
      Eigen::VectorXf U;
      U.resize(trans.persons);
   //fill in U for variance
   for(int p=0;p<trans.persons;p++)
   {
   U[p] = residuals_original[p] *(transGeno[p] - tmp.dot(trans.UDX[p]));
   }
   double var = U.SumSquares();
    */
   if(denominator==0) chisq=0.0;
   else chisq = numerator*numerator/denominator;
   pvalue = pchisq(chisq, 1,0,0);
   //printf("Lin Pvalue is: %g\n",pvalue);
   //pvalue = chidist(chisq, 1);
   effSize = numerator/denominator;
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      //effSize /= (SNPCount*1/)
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);

	 fprintf(out,"\t%s\t%g\n",eff.c_str(),pvalue);
	 fflush(out);
      }
      else
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,eff.c_str(),pvalue);
	 fflush(out);
      }
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);

	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
	 fflush(out);
      }
      else
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,effSize,pvalue);
	 fflush(out);
      }
   }
}

void Burden::MerlinLRT(double L0,FastTransform & fullTrans,FILE * out, Eigen::VectorXf & transGeno,String & geneName,FastFit & fullEngine,String & method)
{
   double L1,chisq,effSize,pvalue;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      L1=fullEngine.EvaluateMulti(fullTrans,false);

      chisq = 2.0*(L1-L0);
      if(chisq<0)
	 chisq = -chisq;
      //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
      pvalue = pchisq(chisq, 1,0,0);
      //pvalue = chidist(chisq, 1);
      effSize = fullEngine.betaHat[fullTrans.fixedEffects-1];
   }
   else 
   {
      for(int row=0;row<fullTrans.persons;row++)
      {
	 fullTrans.UX[row][fullTrans.fixedEffects-1] =transGeno[row];
      }
      fullEngine.OptimizeDelta(0.0001,fullTrans);
      //fullEngine.deltaHat = fullEngine.min;
      fullEngine.GetBeta(fullEngine.deltaHat,fullTrans);
      L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);

      chisq = 2.0*(L1-L0);
      //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
      if(chisq<0)
	 chisq = -chisq;
      //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
      pvalue = pchisq(chisq,1,0,0);
      //pvalue = chidist(chisq,1);
      effSize = fullEngine.beta[fullTrans.fixedEffects-1];
   }
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%s\t%g\n",eff.c_str(),pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,effSize,pvalue);
   }
}

//Genotype here should be the transformed genotype
void Burden::WaldAssoc(FastTransform & fullTrans,FILE * out, Eigen::VectorXf & genotype,String & geneName,FastFit & fullEngine,String & method)
{
   double pvalue,chisq,effSize;
   if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
   {
      fullTrans.UDX.resize(fullTrans.persons,fullTrans.fixedEffects);
      fullTrans.UDY.resize(fullTrans.persons);
      fullTrans.inv.resize(fullTrans.fixedEffects,fullTrans.fixedEffects);

      //fullTrans.FinalizeTransform(0.0,0.0,true,ped);
      effSize = fullEngine.betaHat[fullTrans.fixedEffects-1];
      chisq = effSize*effSize/fullTrans.inv(fullTrans.fixedEffects-1,fullTrans.fixedEffects-1);
      pvalue = pchisq(chisq,1,0,0);
      //pvalue = chidist(chisq,1);
   }
   else
   {
      for(int row=0;row<fullTrans.persons;row++)
      {
	 fullTrans.UX[row][fullTrans.fixedEffects-1] = genotype[row];
      }
      fullEngine.OptimizeDelta(0.0001,fullTrans);
      //fullEngine.deltaHat = fullEngine.min;
      fullEngine.GetBeta(fullEngine.deltaHat,fullTrans);
      effSize = fullEngine.beta[fullTrans.fixedEffects-1];
      fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
      fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;

      fullTrans.UDX.resize(fullTrans.persons,fullTrans.fixedEffects);
      fullTrans.UDY.resize(fullTrans.persons);
      fullTrans.inv.resize(fullTrans.fixedEffects,fullTrans.fixedEffects);

      //fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
      chisq = effSize*effSize/fullTrans.inv(fullTrans.fixedEffects-1,fullTrans.fixedEffects-1);
      pvalue = pchisq(chisq,1,0,0); 
      //pvalue = chidist(chisq,1); 
   }
   //double pvalue = 2.0*tdist(t0, fullTrans.persons-fullTrans.fixedEffects);
   if(method == "MB")
   {
      String eff; 
      eff = "/sqrt(pq)";
      if(fullResult)
      {    
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%s\t%g\n",eff.c_str(),pvalue);
      }    
      else 
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {    
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }    
      else 
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_se,effSize,pvalue);
   }
}

void Burden::CalculateSingleVarPvalue(FastTransform & trans,String & geneName,FastFit & engine,Eigen::MatrixXf & genotypeSKAT,Pedigree & ped)
{
   //printf("Working on singlevar on gene %s.\n",geneName.c_str());
   Eigen::MatrixXf transGeno;
   if(!FastFit::unrelated)
   {
      if(!trans.empKin)
      {
	 transGeno.resize(genotypeSKAT.rows(),genotypeSKAT.cols());
	 transGeno.setZero();
	 for(int m=0;m<genotypeSKAT.cols();m++) 
	 {
	    int index=0;
	    for (int f = 0; f < ped.familyCount; f++)
	    {
	       if (trans.pPheno[f].Length() > 0)
	       {
		  int count = trans.pPheno[f].Length();
		  int block = index;
		  for(int i=0;i<count;i++)
		  {
		     for(int c=0;c<count;c++)
		     {
			transGeno(index,m) += trans.UD(block+i,block+c)*genotypeSKAT(block+c,m);
		     }
		     index++;
		  }
	       }
	    }
	 }
      } else transGeno = trans.UD*genotypeSKAT;
   }
   // Calculate singlevariant pvalues
   singlevarPvalue.resize(genotypeSKAT.cols());
   singlevarPvalue.setZero();
   singlevarEff.resize(genotypeSKAT.cols());
   singlevarEff.setZero();

   Eigen::VectorXf denominator_2;
   denominator_2.resize(trans.fixedEffects);

   Eigen::VectorXf tmp; //tmp is T(sum_SiZi)*inv
   tmp.resize(trans.fixedEffects);
   for(int m=0;m<genotypeSKAT.cols();m++)
   {
      double numerator=0.0, denominator=0.0;
      if(!FastFit::unrelated)
      {
	 denominator_2.setZero();

	 for(int i=0;i<trans.persons;i++)
	 {
	    numerator += residuals_original[i]*transGeno(i,m);
	    denominator += transGeno(i,m)*transGeno(i,m);
	    for(int j=0;j<trans.fixedEffects;j++)
	    {
	       denominator_2[j] += transGeno(i,m)*trans.UDX(i,j);
	    }
	 }

	 tmp= trans.inv* denominator_2;

	 denominator -= tmp.dot(denominator_2);
	 denominator *= sigma2_hat;
      }
      else
      {
	 denominator_2.setZero();

	 for(int i=0;i<trans.persons;i++)
	 {
	    numerator += residuals_original[i]*genotypeSKAT(i,m);
	    denominator += genotypeSKAT(i,m)*genotypeSKAT(i,m);
	    for(int j=0;j<trans.fixedEffects;j++)
	    {
	       denominator_2[j] += transGeno(i,m)*trans.UDX(i,j);
	    }
	 }

	 tmp= trans.inv* denominator_2;

	 denominator -= tmp.dot(denominator_2);
	 denominator *= sigma2_hat;
      }

      double pvalue =  pchisq(numerator*numerator/denominator,1,0,0);
      //double pvalue =  chidist(numerator*numerator/denominator,1);
      singlevarPvalue[m] = pvalue;
      singlevarEff[m] = numerator/denominator;
   }
   //printf("finished working on singlevar on gene %s.\n",geneName.c_str());
}

void Burden::SKATScoreAssoc(FastTransform & trans,FILE * out,String & geneName,FastFit & engine,Pedigree & ped,Eigen::MatrixXf & transGeno,String & method)
{
   //delete the columms that are redundant
   double davies_pvalue,liu_pvalue,original_Qstat;
   bool disect=false;
   double Qstat,pvalue;

   Eigen::MatrixXf GWG;
   GWG.resize(trans.persons,trans.persons);
   GWG.setZero();
   Eigen::VectorXf weight;
   weight.resize(genotypeSKAT.cols());

   for(int i=0;i<weight.size();i++)
   {
      if(method == "uw")
	 weight[i] = 1.0; 

      if(method == "mb")
	 weight[i] = 1.0/sqrt(mafIncluded[i]*(1.0-mafIncluded[i]));

      if(method=="beta")
      {
	 double alpha = 1.0;
	 double beta=25.0;
	 double beta_density = GetBetaDensity(alpha,beta,mafIncluded[i]);
	 weight[i] = beta_density * beta_density;
      }
   }

   /*
      for(int i=0;i<trans.persons;i++)
      {
      for(int m=0;m<genotypeSKAT.cols();m++)
      {
      GWG(i,i) += weight[m]*transGeno(i,m)*transGeno(i,m);
      }
      for(int j=i+1;j<trans.persons;j++)
      {
      for(int m=0;m<genotypeSKAT.cols();m++)
      {
      GWG(i,j) += transGeno(i,m)*transGeno(j,m)*weight[m];
      }
      GWG(j,i) = GWG(i,j);
      }
      }
    */
   Eigen::MatrixXf weight_mat;
   weight_mat.resize(weight.size(),weight.size());
   for(int i=0;i<weight.size();i++)
   {
      for(int j=0;j<weight.size();j++)
      {
	 if(i==j)
	    weight_mat(i,j) = weight[i];
	 else 
	    weight_mat(i,j) = 0.0;
      }
   }

   GWG = transGeno*weight_mat*transGeno.transpose();
   Eigen::VectorXf tmp;
   tmp.resize(trans.persons);
   tmp = residuals_original.transpose()*GWG;
   float result = tmp.dot(residuals_original);
   Qstat = result;
   original_Qstat = Qstat;
   /*
      printf("Qstat is %g\n",Qstat);
      printf("weights are:\n");
      for(int i=0;i<weight.size();i++)
      printf("%g ",weight[i]);
      printf("\n");
    */
   int n = weight.size();
   double * lambda = new double [n];

   CalculateLambda(weight,lambda,transGeno); //now lambda is filled in
   pvalue = MixChidist(lambda, n, Qstat,"Davies");
   liu_pvalue = MixChidist(lambda, n, Qstat,"Liu");
   int disect_itr=0;
   while( (pvalue<=0 ||pvalue==2.0 || isnan(pvalue)) && disect_itr<1000)
   {
      disect=true;
      Qstat*=0.9999;
      pvalue = MixChidist(lambda, n, Qstat,"Davies");
      disect_itr++;
   }
   disect_itr=0;
   while((pvalue<=0 ||pvalue==2.0 || isnan(pvalue))  && disect_itr<1000)
   {
      Qstat*=0.99;
      pvalue = MixChidist(lambda, n, Qstat,"Davies");
      disect_itr++;
   }
   davies_pvalue = pvalue;
   if(lambda) delete [] lambda;
   if(disect)
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t<%g\t%g\n",original_Qstat,davies_pvalue,liu_pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t<%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,original_Qstat,davies_pvalue,liu_pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\t%g\n",original_Qstat,davies_pvalue,liu_pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,original_Qstat,davies_pvalue,liu_pvalue);
   }
}

double Burden::GetBetaDensity(double a, double b, double x)
{
   double density;
   density = dbeta(x,a,b,0);
   //density = exp(gammln(a+b)-gammln(a)-gammln(b)+(a-1.0)*log(x)+(b-1.0)*log(1.0-x));
   return density;
}

void Burden::CalculateLambda(Eigen::VectorXf & weight,double lambda [],Eigen::MatrixXf & transGeno)
{
   //Goal of this function is to fill in lambda
   //now we only take diagnol weight matrix 
   Eigen::MatrixXf WG;
   WG.resize(genotypeSKAT.cols(),genotypeSKAT.rows());
   for(int i=0;i<WG.rows();i++)
   {
      for(int j=0;j<WG.cols();j++)
      {
	 WG(i,j) = transGeno(j,i) * sqrt(weight[i]);	
      }
   }
   Eigen::MatrixXf final = WG*P0*WG.transpose();
   //use EIGEN library
   Eigen::JacobiSVD<Eigen::MatrixXf> svd(final, Eigen::ComputeThinU);

   const Eigen::VectorXf S = svd.singularValues();

   for(int i=0;i<weight.size();i++)
   {
      lambda[i] = S[i];
   }
}

void Burden::MerlinFastAssoc(FastTransform & trans,FILE * out, Eigen::VectorXf & transGeno,String & geneName,FastFit & engine,String & method)
{
   //printf("Now working on %s in gene %s\n",method.c_str(),geneName.c_str());
   double numerator = 0.0;
   double denominator = 0.0;

   numerator = residuals_original.dot(transGeno);
   denominator = transGeno.dot(transGeno);
   double chisq,effSize,pvalue;
   //printf("numerator and denominator are: %g, %g\n",numerator,denominator);

   effSize = numerator/denominator;

   if(FastFit::unrelated)
   {
      denominator *= engine.sigma2Hat;
   }

   if(denominator==0) chisq=0.0;
   else chisq = numerator*numerator/denominator;
   pvalue = pchisq(chisq,1,0,0);
   //pvalue = chidist(chisq,1);
   //Adjust effSize to make sense
   if(method == "MB")
   {
      String eff;
      eff = "/sqrt(pq)";
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g%s\t%g\n",effSize,eff.c_str(),pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g%s\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,effSize,eff.c_str(),pvalue);
   }
   else
   {
      if(fullResult)
      {
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%s\t",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,varsIncluded.c_str());
	 for(int m=0;m<mafIncluded.Length();m++)
	    fprintf(out,"%g,",mafIncluded[m]);
	 fprintf(out,"\t");
	 for(int m=0;m<singlevarPvalue.size();m++)
	    fprintf(out,"%g,",singlevarPvalue[m]);

	 fprintf(out,"\t");
	 for(int m=0;m<singlevarEff.size();m++)
	    fprintf(out,"%g,",singlevarEff[m]);
	 fprintf(out,"\t%g\t%g\n",effSize,pvalue);
      }
      else
	 fprintf(out,"%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%g\t%g\n",geneName.c_str(),numVarPerGene,mac,freq_mean*numVarPerGene*2.0,freq_min,freq_max,freq_mean,freq_se,effSize,pvalue);
   }
}

void Burden::PrintFooter(FILE * output,int bad,GroupFromAnnotation & group)
{
   fprintf(output,"===================================================================\n");
   fprintf(output,"\n%d %s tested.\n",group.geneCount,group.geneCount>1?"genes were":"gene was");
   fprintf(output,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");
}

void Burden::PrintHeader(FILE * output,FastTransform & trans,FastFit & engine,String & traitName)
{
   double h2 = 100.0*engine.sigma_g2Hat/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv = 100.0* engine.sigma_s2Hat/trans.traitVar;

   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, Xheritability = %6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }

   fprintf(output,"===================================================================\n");

   if(fullResult)
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tVarsIncluded\tMafIncluded\tSingleVarPvalue\tSingleVarEffSize\tEffSIZE\tpvalue\n");
   else
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tEffSIZE\tpvalue\n");
}
void Burden::PrintSKATHeader(FILE * output,FastTransform & trans,FastFit & engine,String & traitName)
{
   double h2 = 100.0*engine.sigma_g2Hat/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv = 100.0* engine.sigma_s2Hat/trans.traitVar;

   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, Xheritability = %6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.4f, variance=%6.4f, heritability=%6.4f, SharedEnv = %6.4f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }

   fprintf(output,"===================================================================\n");
   if(fullResult)
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tVarsIncluded\tMafsIncluded\tSingleVarPvalue\tSingleVarEffSize\tQstat\tDaviesPvalue\tLiuPvalue\n");
   else
      fprintf(output,"Gene\tNumVarInGene\tInformative_AC\tAvgBurden\tFreqMin\tFreqMax\tFreqMean\tFreqSE\tQstat\tDaviesPvalue\tLiuPvalue\n");
}

